/**
 * Point light source (with just ambient color)
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.11.11
 */
#ifndef LIGHT_POINT_H
#define LIGHT_POINT_H

#include "RayPrimitive.h"
#include "Vector.h"


/// basic definition of a point light source
class ltPoint {

	public:
		/**
		 * Basic constructor
		 * @param loc - location of the light
		 * @param clr - color of this light
		 */
		ltPoint( const Vector &loc=Vector(0.f,0.f,0.f), const Color &clr=Color(1.f,1.f,1.f) )
			: m_clrAmbient(clr), m_vLocation(loc) {}


		/**
		 * Constructor which knows how to load point location
		 *   from the TRaX main memory
		 * @param memAddress - memory address for the data
		 * @param clr        - color of the light
		 */
		ltPoint( const int &memAddress, const Color &clr=Color(1.f,1.f,1.f) ) : m_clrAmbient(clr) {
			m_vLocation.set( memAddress );
		}


	// public for speed
		Vector m_vLocation;			/// location of this point light
		Color m_clrAmbient;			/// ambient color of this light
};


#endif // LIGHT_POINT_H
