/**
 * This is the main file pulling together all the bits necessary for 
 *   CS 6965 Project 4 - using BVH
 *
 * This is the extension of no inheritance project 2
 *
 * @author:	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	10.06.11
 */

#define DO_REFLECTIONS		// flag whether to compile with reflection-enabling code bits (decreases speed)
//#define DO_MULTIOBJECT		// flag whether to keep track of multiple objects (decreases speed)
#include "project4.h"

/**
 * Which scene do we compile to run? -> soon i'll do a generate scene file or something...
 * 0 - required image
 */
#define TRAX_SCENE 0
#define SCENE TRAX_SCENE

#define BVH_TRAVERSAL_STACK_SIZE 32			// size of the stack used to queue up bvh boxes to go through
#define DO_BVH 1		// flag whether to do BVH traversal OR all triangles

/// main code to do the entire raytracing image :D
int main() {

	// Setup code. On the CPP target sets up global memory.
	trax_setup();

	// Load the pointer to the frame buffer in global memory
	int framebuffer = GetFrameBuffer();
	int widthINT	= loadi( 0, 1 );
	float widthInv	= loadf( 0, 2 );
	float width		= loadf( 0, 3 );
	int heightINT	= loadi( 0, 4 );
	float heightInv	= loadf( 0, 5 );
	float height	= loadf( 0, 6 );


	// set up Required scenes
#if SCENE==TRAX_SCENE		// scene 8 = required image

	// set up things we know in advance, without loading scene file
	const Color ltAmb( 0.4f, 0.4f, 0.4f );
	const float kDiff = 0.7f;
	const float kAmb  = 0.3f;

	// load BVH address
	const int sAddrBVH = loadi( 0, 8 );

	// load 1st material (to try)
	const int sAddrMat = loadi( 0, 9 );

	// load camera -> computes aspect ration and such on its own
	myCamera cam( loadi(0, 10) );

	// load background color
	const Color backgroundC( loadi(0, 11) );

	// load point light
	ltPoint ptLt( loadi(0, 12), Color(1.f,1.f,1.f) );

	// load triangle information
	const int sAddrTri = loadi( 0, 28 );
	const int numTri   = loadi( 0, 29 );

#endif


	// set up scene construct
	RayPrimitive ray( cam.getEye() );
	const float iPrem = 2.f * widthInv;
	const float iSub  = 1.f - widthInv;			// to make things match -> need to switch - to +
	const float jPrem = 2.f * heightInv;
	const float jSub  = 1.f - heightInv;


	// Iterate over all of the pixels
	// Using atomicinc here allows your code to have only one thread execute per return value
	for( int pixel = atomicinc(0); pixel < width * height; pixel = atomicinc(0) ) {

		// Get screen location
		int i = pixel % heightINT;
		int j = pixel / widthINT;

		// 1. My camera
		cam.generateRayDirection( ray, i*iPrem - iSub, j*jPrem - jSub );


		// 2. intersect with intersection records
		IntersectionRecord closestIr;
		ray.m_fMinT = MYRTC::RAY_EPS;
		ray.m_fMaxT = MYRTC::RAY_INF;

	#if DO_BVH==0
		gpTriangle tmpT;
		int tID			 = 0;
		bool intersected = false;
		for( int tAddr=sAddrTri; tID<numTri; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
			tmpT.set( tAddr );
			intersected |= tmpT.intersectDetailed( ray, closestIr );
		}

		// 3. shade point but only if on the correct side
		if( intersected ) {
	#else
		// 3. shade point
		if( intersectDetailedWithBVH( ray, closestIr, sAddrBVH ) ) {
	#endif

			// ambient light computation
			Color outClr( ltAmb );
		#if SCENE==TRAX_SCENE	// TRaX
			//MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID, kDiff, kAmb );
			//outClr.mult( tmpM.m_fKamb );
			outClr.mult( kAmb );		// for now, we can save some cycles
		#else
			outClr.mult( closestIr.m_pGeomMat->m_fKamb );
		#endif

			// get intersection point/normal
			Vector ptLoc, ptNorm;
			ray.getPoint( ptLoc, ray.m_fMaxT );

			gpTriangle tmpT;
			tmpT.set( closestIr.m_iGeomHitAddr );
			closestIr.m_iGeomMatID = tmpT.m_matID;
			tmpT.getNormal( ptNorm, ptLoc );

			RayPrimitive shRay( ptLoc );
			shRay.m_fMinT = MYRTC::RAY_INTER_STEP;

			// look through all lights in scene and accumulate shading
		#if SCENE==TRAX_SCENE	// TRaX
			{	// get shadow ray to light
				Vector intToLight( ptLt.m_vLocation );
		#else
			for( unsigned int l=0; l<numPtLights; ++l ) {
				// get shadow ray to light
				Vector intToLight( ptLts[l]->m_vLocation );
		#endif
				intToLight.sub( shRay.m_vOrigin );
				shRay.m_vDirection.set( intToLight );
				shRay.m_fMaxT = 1.0f;

				// intersect shadow ray with scene -> slower
			#if DO_BVH==0
				bool outShadow  = true;
				int tID			= 0;
				for( int tAddr=sAddrTri; tID<numTri && outShadow; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.set( tAddr );
					outShadow = tmpT.intersectQuick( shRay );
				}

				// do we need to shade?
				if( outShadow ) {
			#else
				if( intersectQuickWithBVH( shRay, sAddrBVH ) ) {
			#endif

					// get cos-varied amount of light reaching the point (for some reason backfacing was an issue :( )
					//const float cosPhi = max( ptNorm.dot( shRay.m_vDirection ) * invsqrt( shRay.m_vDirection.lengthSq() ), 0.f );
					const float normDot = ptNorm.dot( shRay.m_vDirection );
					const float cosPhi = (normDot > 0.f ? normDot : -normDot ) * invsqrt( shRay.m_vDirection.lengthSq() );

					// light we get can be seen -> do lambertian (diffuse shading)
				#if SCENE==TRAX_SCENE // TRAX
					Color tmpClr( ptLt.m_clrAmbient );
					//tmpClr.mult( (cosPhi*tmpM.getKDiff() ) );
					// for now can save some cycles
					tmpClr.mult( (cosPhi*kDiff) );
				#else
					Color tmpClr( ptLts[l]->m_clrAmbient );
					tmpClr.mult( (cosPhi*closestIr.m_pGeomMat->m_fKdiff) );
				#endif
					outClr.add( tmpClr );
				}

			} // end l

			// apply diffuse material color
		#if SCENE==TRAX_SCENE
			MatPrimitive tmpM( sAddrMat, closestIr.m_iGeomMatID );
			outClr.mult( tmpM.m_clrDiff );
		#else
			outClr.mult( closestIr.m_pGeomMat->m_clrDiff );
		#endif

			outClr.clamp01();
			StorePixelIntoFB( framebuffer, pixel, outClr );
			continue;
		} // end intersected

		// background color
		StorePixelIntoFB( framebuffer, pixel, backgroundC );
	}

	// Conditional to only execute the following on the CPU version
#if TRAX==0
	printf( "Done drawing.\n" );
#endif

	// Cleanup code. On the CPP target writes out.png.
	trax_cleanup();
}




/// Helper function that helps us to store a pixel into the frame buffer
void StorePixelIntoFB( const int &fb, const int &pixel, const Color &clr ) {
	int base = fb + pixel*3;
	storef( clr.r, base, 0 );
	storef( clr.g, base, 1 );
	storef( clr.b, base, 2 );
}


/**
 * Helper function that throws a ray through our scene and traverses
 *   the BVH, while keeping track of the intersection record
 * NOTE: this works only for the TRaX triangle-only scenes
 *
 * @param ray    - ray to cast through the BVH (will be updated with far t-value)
 * @param intRec - record holding intersection information
 */
inline bool intersectDetailedWithBVH( RayPrimitive &ray, IntersectionRecord &intRec, const int &sAddrBVH ) {

	int bvhStack[BVH_TRAVERSAL_STACK_SIZE];
	ray.computeBoxIntersectInfo();
	
	bool intersected = false;
	int stackOffset  = 0, nodeId = 0;
	float t0, t1;
	
	gpBox      tmpB;
	gpTriangle tmpT;

	while( true ) {

		// load BVH box
		int tmpAddr = sAddrBVH + gpBox::SIZE_IN_TRAX*nodeId;
		tmpB.set( tmpAddr );

		// intersect with ray
		if( tmpB.intersectDetailed( t0, t1, ray ) ) {

			// are we a parent node?
			int numChildren = loadi( tmpAddr, 6 );
			if( numChildren < 0 ) {
			/*	nodeId					= loadi( tmpAddr, 7 );	// traverse left child
				bvhStack[stackOffset++] = nodeId+1;				// push right child
				continue;
			//*/
				// check which child to add first based on which is closer
				// left is closer
				if( ( ray.m_iDirSignX && tmpB.pts[1].x>loadf(tmpAddr,0) ) || 
					( ray.m_iDirSignY && tmpB.pts[1].y>loadf(tmpAddr,1) ) ||
					( ray.m_iDirSignZ && tmpB.pts[1].z>loadf(tmpAddr,2) ) ) {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId+1;
					continue;
				}
				// right is closer
				else {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId++;
					continue;
				}
			}

			// are we leaf?
			else {

				//intRec.m_iGeomHitAddr = nodeId;
				//return true;

				// intersect with all interior children
				int tAddr = loadi( tmpAddr, 7 );
				int tID	  = 0;
				for( ; tID<numChildren; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.setQuick( tAddr );
					intersected |= tmpT.intersectDetailed( ray, intRec );
				}

				// pop current box off stack
//				if( stackOffset==0 )
//					break;
//				nodeId = bvhStack[--stackOffset];
			}
		}
//		else {

			// no intersection -> pop current box off stack
			if( stackOffset==0 )
				break;
			nodeId = bvhStack[--stackOffset];
//		}
	}
	/**/
	return intersected;
}


/**
 * Quick intersection of a ray with the BVH scene. This is intended for 
 *   shadow rays only. Returns whether a ray has hit anything
 * NOTE: this works only for the TRaX triangle-only scenes
 *
 * @param ray - ray to cast through the BVH (will be updated with far t-value)
 * @return flag whether ray hits anything
 */
inline bool intersectQuickWithBVH( RayPrimitive &ray, const int &sAddrBVH ) {

	int bvhStack[BVH_TRAVERSAL_STACK_SIZE];
	ray.computeBoxIntersectInfo();
	
	int stackOffset = 0, nodeId = 0;
	float t0, t1;
	
	gpBox      tmpB;
	gpTriangle tmpT;

	while( true ) {

		// load BVH box
		int tmpAddr = sAddrBVH + gpBox::SIZE_IN_TRAX*nodeId;
		tmpB.set( tmpAddr );

		// intersect with ray
		if( tmpB.intersectDetailed( t0, t1, ray ) ) {

			// are we a parent node?
			int numChildren = loadi( tmpAddr, 6 );
			if( numChildren < 0 ) {
			/*	nodeId					= loadi( tmpAddr, 7 );	// traverse left child
				bvhStack[stackOffset++] = nodeId+1;				// push right child
				continue;
			*/
				// check which child to add first based on which is closer
				// left is closer
				if( ( ray.m_iDirSignX && tmpB.pts[1].x>loadf(tmpAddr,0) ) || 
					( ray.m_iDirSignY && tmpB.pts[1].y>loadf(tmpAddr,1) ) ||
					( ray.m_iDirSignZ && tmpB.pts[1].z>loadf(tmpAddr,2) ) ) {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId+1;
					continue;
				}
				// right is closer
				else {
					nodeId					= loadi( tmpAddr, 7 );
					bvhStack[stackOffset++] = nodeId++;
					continue;
				}
			}

			// are we leaf?
			else {

				// intersect with all interior children
				int tAddr = loadi( tmpAddr, 7 );
				int tID	  = 0;
				for( ; tID<numChildren; ++tID, tAddr+=gpTriangle::SIZE_IN_TRAX ) {
					tmpT.setQuick( tAddr );
					if( tmpT.intersectQuick( ray )==false )
						return false;
				}

				// pop current box off stack
//				if( stackOffset==0 )
//					break;
//				nodeId = bvhStack[--stackOffset];
			}
		}
//		else {

			// no intersection -> pop current box off stack
			if( stackOffset==0 )
				break;
			nodeId = bvhStack[--stackOffset];
//		}
	}
	/**/
	return true;
}
